import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"

os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
import numpy as np
import scipy
import logging
from xframe import log
from xframe.startup_routines import load_recipes
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_trapz
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_ht
from xframe import settings
from xframe import Multiprocessing
import xframe
xframe.settings.general.n_control_workers = 1
xframe.controller.control_worker.restart_working()
#log=log.setup_custom_logger('root','INFO')

analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
n_radial_points = 256
max_order = 99
orders = np.arange(max_order+1)
r_max = 1000


#wd3 = generate_weightDict_zernike(max_order,n_radial_points,n_cpus = 30,dimensions=3)
#wd33 = generate_weightDict_trapz(max_order,n_radial_points,n_cpus = 30,dimensions=3)
#w3 = wd3['weights']
#w33 = wd33['weights']
#
#wd2 = generate_weightDict_zernike(max_order,n_radial_points,n_cpus = 30,dimensions=2)
#wd22 = generate_weightDict_trapz(max_order,n_radial_points,n_cpus = 30,dimensions=2)
#w2 = wd2['weights']
#w22 = wd22['weights']

wd2 = generate_weightDict_trapz(max_order,n_radial_points,n_cpus = 30,dimensions = 2)
ht2,iht2 = generate_ht(wd2['weights'],orders,r_max,pi_in_q=True,dimensions = 2)

wd3 = generate_weightDict_trapz(max_order,n_radial_points,n_cpus = 30,dimensions = 3)
ht3,iht3 = generate_ht(wd3['weights'],orders,r_max,pi_in_q=True,dimensions = 3)
ht3g,iht3g = generate_ht(wd3['weights'],orders,r_max,pi_in_q=True,dimensions = 3,use_gpu=True)
#
d2 = np.ones((n_radial_points,2*len(orders)-1))
i2 = iht2(ht2(d2))
diff2 = i2-d2
rd2 = np.mean(np.sqrt((diff2*diff2.conj()).real)/d2)
#
m_orders=np.concatenate((np.arange(max_order+1,dtype = int),-np.arange(max_order,0,-1,dtype = int)))
d3 = [ np.ones((n_radial_points,max_order+1)) for m in m_orders]
nlm = max_order*(max_order+2)+1
d3g = np.ones((n_radial_points,nlm),dtype = complex)
i3 = iht3(ht3(d3))
i3g = iht3g(ht3g(d3g))


diff3g= d3g-i3g
rd3g = np.mean(np.sqrt((diff3g*diff3g.conj()).real)/d3g)


#wd3 = generate_weightDict_zernike(max_order,n_radial_points,n_cpus = 30)
#w3 = assemble_weights_zernike(wd3['weights'],np.arange(100),1000,True)
#
#w3f = w3['forward']
#w3i = w3['inverse']
#
#d3f = np.sum(w3f*10,axis = 0)
#d3i = np.sum(d3f[1:,None,:]*w3i,axis=0)
#
#weight_dict2 = generate_weightDict_zernike(max_order,n_radial_points,n_cpus = 30,dimension=2)
#w2 = assemble_weights_zernike(weight_dict2['weights'],np.arange(100),1000,True,dimension=2)
##
#w2f = w2['forward']
#w2i = w2['inverse']
##
#d2f = np.sum(w2f*10,axis = 0)
#d2i = np.sum(d2f[1:,None,:]*w2i,axis=0)
#
#ww_pi = calc_zernike_weights_pi_2(np.arange(100),500,128)
#ww = calc_zernike_weights_2(np.arange(100),500,128)
#zht,izht = generate_zernike_ht_old(ww,np.arange(100),1000)

#wd2 = np.moveaxis(calc_zernike_weights_pi_2(np.arange(100),500,128),0,2)
#d2f2 = np.sum(wd2,axis = 0)
#d2i2 = np.sum(d2f2[:,None,:]*wd2,axis = 0)



import numpy as np
import numpy.ma as mp
import scipy.integrate as spIntegrate
from scipy.special import jv as bessel_jv
from scipy.special import roots_legendre
import logging
from itertools import repeat

log=logging.getLogger('root')

from xframe.library.mathLibrary import eval_ND_zernike_polynomials
from xframe.library.mathLibrary import bessel_jnu
from xframe.library.mathLibrary import spherical_bessel_jnu
from xframe.library.mathLibrary import polar_spherical_dft_reciprocity_relation_radial_cutoffs
from xframe import Multiprocessing
#######################################################
###    polar/spherical HT via Zernike expansion     ###

def generate_weightDict_zernike(max_order, n_radial_points,expansion_limit = -1,pi_in_q = True,dimensions=3,n_cpus=False):
    '''
    Calculates the weights for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    Strangly the last sentence does not seem to be right, the transform gets better for higher limits thats why, quite arbitrarily, the default below corresponds to 2*(2*n_radial_points-1)
    '''
    orders = np.arange(max_order+1)
    assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    
    if expansion_limit < 0:
        expansion_limit = 2*(2*n_radial_points - 1)
        log.info('Default zernike expansion limit = {}'.format(expansion_limit))
    expansion_limit = max(expansion_limit,max_order)

    worker_by_dimensions= {3:calc_spherical_zernike_weights,2:calc_polar_zernike_weights}
    worker = worker_by_dimensions[dimensions]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,argArrays=[orders],const_args=[n_radial_points,expansion_limit,pi_in_q],callWithMultipleArguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,argArrays=[orders],const_args=[n_radial_points,expansion_limit,pi_in_q],callWithMultipleArguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'Zerinike_{}D'.format(dimensions)
    return weightDict


def generate_weightDict_trapz(max_order, n_radial_points,pi_in_q = True,dimensions=3,n_cpus=False):
    '''
    Calculates the weights for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    Strangly the last sentence does not seem to be right, the transform gets better for higher limits thats why, quite arbitrarily, the default below corresponds to 2*(2*n_radial_points-1)
    '''
    orders = np.arange(max_order+1)
    assert 2*n_radial_points-1>max_order,'Not enough radial points (N_r={}) to calculate weights upto order {}. max_order<2*n_radial_points-1 '.format(n_radial_points,max_order)
    

    worker_by_dimensions= {2:calc_polar_trapz_weights,3:calc_spherical_trapz_weights}
    worker = worker_by_dimensions[dimensions]
    if not isinstance(n_cpus,bool):
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,argArrays=[orders],const_args=[n_radial_points,pi_in_q],callWithMultipleArguments=True,n_processes=n_cpus)
    else:
        weights = Multiprocessing.comm_module.request_mp_evaluation(worker,argArrays=[orders],const_args=[n_radial_points,pi_in_q],callWithMultipleArguments=True)
    weightDict={}
    weightDict['weights'] = weights
    weightDict['posHarmOrders'] = orders
    weightDict['mode'] = 'Zerinike_{}D'.format(dimensions)
    return weightDict

# this is equivalent to trapz for expansion_limit -> infinity -.- so no good
def calc_spherical_zernike_weights(orders,n_radial_points,expansion_limit,pi_in_q,**kwargs):
    '''
    Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
    There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
    $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    
    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(1,radial_points)
    n_p = radial_points-1 
    ks=np.arange(radial_points)
    if pi_in_q:
        j_factor=np.pi        
    else:
        j_factor=1/2
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(l):
        zernike_dict=eval_ND_zernike_polynomials(np.array([l]),expansion_limit,ps/radial_points,3)
        #log.info('finished_zernike dict')
        s=np.arange(l,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-l)/2 )*(2*s+3)
        jp=spherical_bessel_jnu(np.repeat((s+1)[:,None],radial_points-1,axis=1),ks[1:]*j_factor)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zk=zernike_dict[l]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        log.info('making summands for order = {}'.format(l))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,1:] = prefactor[:,None,None]*Zk[:,:,None]*jp[:,None,:]
        if l == 0:
            summands[0,:,0] = j_factor
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,1:]=np.square(ps)[:,None]/(ks[None,1:])
    c_kp[:,0]=np.square(ps)
    weights=weights*c_kp[None,:,:]
    return weights

def calc_spherical_trapz_weights(orders,n_radial_points,pi_in_q,**kwargs):
    '''
    Generates weights for trapezooidal rule
    '''
    N = n_radial_points
    ps=np.arange(1,N)
    ks=np.arange(N)
    if pi_in_q:
        j_factor=np.pi        
    else:
        j_factor=1/2

    ls = orders
    jmpk = spherical_bessel_jnu(np.repeat(np.repeat(ls[:,None,None],N-1,axis=1),N,axis=2),ks[None,:]*ps[:,None]*j_factor/N)
    weights = ps[None,:,None]**2*jmpk
    return weights


# this is equivalent to trapz for expansion_limit -> infinity -.- so no good
def calc_polar_zernike_weights(orders,n_radial_points,expansion_limit,pi_in_q,**kwargs):
    '''
     Calculates the weights $w_{l,p,k}$ for the Zernike version of the approximated hankel transform.
     There is no sense to chose expansion_limit higher than 2*n_radial_points-1, since this is the limit ing degree of a polynomial (i.e. also the Zernike polynomials) that could be resolved on a gauss legendre grid. In this routine we use the trapezoidal rule so it is probably advisible to stay below that limit.
     $l$ indexes the order, $p$ the new_radial_coordinate and  $k$ the summation radial coordinate.
    '''    

    radial_points=n_radial_points    
    #radial_points=opt['n_radial_points']+1
    
    ps=np.arange(1,radial_points)
    n_p = radial_points-1 
    ks=np.arange(radial_points)
    if pi_in_q:
        J_factor=np.pi        
    else:
        J_factor=1/2
    #    log.info('ps={}'.format(ps/2))
    n_k = radial_points
    #weights=np.zeros((len(orders),n_p,n_k))
    #zernike_dict=eval_ND_zernike_polynomials(orders,expansion_limit,ks/radial_points,3)

    def weights_for_fixed_order(m):
        zernike_dict=eval_ND_zernike_polynomials(np.array([m]),expansion_limit,ps/radial_points,2)
        #log.info('finished_zernike dict')
        s=np.arange(m,expansion_limit+1,2)
        len_s=len(s)
        prefactor=(-1)**( (s-m)/2 )*(2*s+2)
        Jk = bessel_jnu(np.repeat((s+1)[:,None],radial_points-1,axis=1),ks[1:]*J_factor)
        #log.info('jp shape =  {}'.format(jp.shape))
        Zp=zernike_dict[m]
        #log.info('Zk shape = {}'.format(Zk.shape))
        # log.info('prefactor = \n {}'.format(prefactor))
        #log.info('Zk = \n {}'.format(Zk))
        log.info('making summands for order = {}'.format(m))
        summands = np.zeros((len_s,n_p,n_k))
        summands[:,:,1:] = prefactor[:,None,None]*Zp[:,:,None]*Jk[:,None,:]
        if m == 0:
            summands[0,:,0] = J_factor
        weights=np.sum(summands,axis=0)
        #log.info('w sum = {}'.format(weights.sum()))#
        return weights
    weights=np.array(tuple(map(weights_for_fixed_order,orders)))
    c_kp=np.zeros((n_p,n_k))
    c_kp[:,1:]=ps[:,None]/(ks[None,1:])
    c_kp[:,0]=ps
    weights=weights*c_kp[None,:,:]
    return weights

def calc_polar_trapz_weights(orders,n_radial_points,pi_in_q,**kwargs):
    '''
    Generates weights for trapezoidal rule.
    '''
    N = n_radial_points
    ps=np.arange(1,N)
    ks=np.arange(N)
    if pi_in_q:
        J_factor=np.pi**2        
    else:
        J_factor=1/2**2
        #    log.info('ps={}'.format(ps/2))
    ms = orders
    Jmpk = bessel_jnu(np.repeat(np.repeat(ms[:,None,None],N-1,axis=1),N,axis=2),ks[None,:]*ps[:,None]*np.pi/N)
    weights = ps[None,:,None]*Jmpk
    return weights



def assemble_weights_zernike(weights,orders,r_max,pi_in_q,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,pi_in_q = pi_in_q)

    if dimensions == 3:
        forward_prefactor = (-1.j)**(orders[None,None,:])*(np.pi**2/q_max**3)*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(np.pi**2/r_max**3)*np.sqrt(2/np.pi)
    elif dimensions == 2:
        all_orders = np.concatenate((orders,orders[:0:-1]))
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(np.pi/q_max**2)
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(np.pi/r_max**2)
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)   
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}

def assemble_weights_trapz(weights,orders,r_max,pi_in_q,dimensions=3):
    '''
    Generates weights for the forward and inverse transform by multiplieing with the propper constants.
    And changes the Array order of the weights from (order,summation_radial_corrdinate,new_radial_coordinate) to 
    (summation_radial_coordinate,new_radial_coordinate,order)
    '''
    n_radial_points = weights.shape[-1]
    q_max=polar_spherical_dft_reciprocity_relation_radial_cutoffs(r_max,n_radial_points,pi_in_q = pi_in_q)

    if dimensions == 2:
        all_orders = np.concatenate((orders,orders[:0:-1]))
        forward_prefactor = (-1.j)**(all_orders[None,None,:])*(np.pi/q_max)**2
        inverse_prefactor = (1.j)**(all_orders[None,None,:])*(np.pi/r_max)**2
        # weights for negative orders are given by $w_{-mpk}=(-1)^m $w_{mpk}$ due to $J_{-m}(x)=(-1)^m J_m(x)$.
        weights = np.concatenate((weights,(-1)**orders[:0:-1,None,None]*weights[:0:-1]),axis = 0)
    elif dimensions == 3:
        forward_prefactor = (-1.j)**(orders[None,None,:])*(np.pi/q_max)**3*np.sqrt(2/np.pi)
        inverse_prefactor = (1.j)**(orders[None,None,:])*(np.pi/r_max)**3*np.sqrt(2/np.pi)
        
    weights=np.moveaxis(weights,0,2)
    forward_weights=weights*forward_prefactor
    inverse_weights=weights*inverse_prefactor
    #log.info('weights shape = {}'.format(forward_weights.shape))
    return {'forward':forward_weights,'inverse':inverse_weights}


def generate_ht(weights,orders,r_max,pi_in_q=False,dimensions=3,use_gpu=False,mode = 'trapz'):
    '''
    Routine that selects the construciton routine for 2D/3D Hankel transforms via zernike expansion. 
    '''
    if mode == 'trapz':
        w = assemble_weights_trapz(weights,orders,r_max,pi_in_q,dimensions=dimensions)
    elif mode == 'zernike':
        w = assemble_weights_zernike(weights,orders,r_max,pi_in_q,dimensions=dimensions)

    spherical_gpu_version = (dimensions == 3) and use_gpu
    spherical_version = (dimensions == 3) and (not use_gpu)
    polar_version = (dimensions == 2)
    if spherical_gpu_version:
        zht,izht = generate_spherical_zernike_ht_gpu(w,orders,r_max,pi_in_q=pi_in_q)
    elif spherical_version:
        zht,izht = generate_spherical_zernike_ht(w,orders,r_max,pi_in_q=pi_in_q)
    elif polar_version:
        zht,izht = generate_polar_zernike_ht(w,orders,r_max,pi_in_q=pi_in_q)
    return zht,izht

def generate_polar_zernike_ht(w,pos_orders,r_max,pi_in_q=False):
    '''
    Generates polar hankel transform using Zernike weights. $HT_m(f_m) = \sum_k f_m(k)*w_kpm$ for $m\geq 0$.
    '''
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_steps = forward_weights.shape[1]
    n_orders = len(pos_orders)
    full_shape = (n_radial_steps,2*n_orders-1)
    out_forward = np.zeros(full_shape,dtype = complex)
    out_inverse = np.zeros(full_shape,dtype = complex)
    def zht(harmonic_coeff):
        reciprocal_harmonic_coeff = np.sum(forward_weights*harmonic_coeff[1:,None,:],axis = 0,out=out_forward)
        return reciprocal_harmonic_coeff
    def izht(reciprocal_coeff):
        harmonic_coeff = np.sum(inverse_weights*reciprocal_coeff[1:,None,:],axis = 0,out=out_inverse)
        return harmonic_coeff
    return zht,izht

def generate_spherical_zernike_ht(w,l_orders,r_max,pi_in_q=False):
    forward_weights= w['forward']
    inverse_weights= w['inverse']

    l_max=l_orders.max()
    m_orders=np.concatenate((np.arange(l_max+1,dtype = int),-np.arange(l_max,0,-1,dtype = int)))
    def zht(harmonic_coeff):
        return tuple(np.sum(forward_weights[:,:,np.abs(m):]*harmonic_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders) 
    def izht(reciprocal_coeff):
        return tuple(np.sum(inverse_weights[:,:,np.abs(m):]*reciprocal_coeff[m][1:,None,:l_max - np.abs(m)+1],axis=0) for m in m_orders)
    return zht,izht

def generate_spherical_zernike_ht_gpu(w,l_orders,r_max,pi_in_q = False):
    forward_weights= w['forward']
    inverse_weights= w['inverse']
    n_radial_points = forward_weights.shape[1]

    l_max=l_orders.max()

    nq = n_radial_points
    nl = len(l_orders)
    nlm = l_max*(l_max+2)+1

    kernel_str = """
    __kernel void
    apply_weights(__global double2* out, 
    __global double2* w, 
    __global double2* rho, 
    long nq,long nlm, long nl)
    {
  
    long i = get_global_id(0); 
    long j = get_global_id(1);
    long l = (long) sqrt((double)j);

 
    // value stores the element that is 
    // computed by the thread
    double2 value = 0;
    // wlm is of shape (sum_q,nq,m) where sum_q = nq-1
    for (int q = 0; q < nq-1; ++q)
    {
    double2 wqql = w[q*nq*nl + i*nl + l];
    double2 rqlm = rho[(q+1)*nlm + j];
    value.x += wqql.x * rqlm.x - wqql.y * rqlm.y;
    value.y += wqql.x * rqlm.y + wqql.y * rqlm.x;
    }
    
    // Write the matrix to device memory each 
    // thread writes one element
    out[i * nlm + j] = value;//w[nq*nl+i*nl + l];
    }
    """

    kernel_dict_forward={
            'kernel': kernel_str,
            'name': 'forward_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [np.complex,np.complex,np.complex,np.int64,np.int64,np.int64],
                'shapes' : [(nq,nlm),forward_weights.shape,(nq,nlm),None,None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input','const_input'],
                'const_inputs' : [None,forward_weights,None,np.long(nq),np.long(nlm),np.long(nl)],
                'global_range' : (nq,nlm),
                'local_range' : None
            }]
        }
    
    kernel_dict_inverse={
            'kernel': kernel_str,
            'name': 'inverse_hankel',
            'functions': [{
                'name': 'apply_weights',
                'dtypes' : [np.complex,np.complex,np.complex,np.int64,np.int64,np.int64],
                'shapes' : [(nq,nlm),inverse_weights.shape,(nq,nlm),None,None,None],
                'arg_roles' : ['output','const_input','input','const_input','const_input','const_input'],
                'const_inputs' : [None,inverse_weights,None,np.long(nq),np.long(nlm),np.long(nl)],
                'global_range' : (nq,nlm),
                'local_range' : None
            }]
        }

    forward_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_forward)
    inverse_gpu_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict_inverse)

    zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(forward_gpu_process)
    inverse_zernike_hankel_transform = Multiprocessing.comm_module.add_gpu_process(inverse_gpu_process)
    return zernike_hankel_transform,inverse_zernike_hankel_transform





import argparse
import importlib
import time,timeit
import profile
import time
import traceback
import os
import sys
from os.path import join as pjoin



def create_project_parsers(parser,known_projects,projects_to_load,default):
   from xframe.startup_routines import DefaultProjectArgparser
   
   if not isinstance(projects_to_load,(list,tuple)):
      projects_to_load = tuple(known_projects.keys())
   subparser = parser.add_subparsers(help='Available Projects',dest='project')
   subparser.default=default
   for project_name,project_path in known_projects.items():
      if project_name in projects_to_load:
         try:
            _argparser_ = importlib.import_module('xframe.projects.'+project_name+'._argparser_')
            #print('alive')
            p = getattr(_argparser_,'ProjectArgparser',DefaultProjectArgparser)()
            #print('alive2')
            p.add_parser(subparser,project_name,project_path)
            #print('alive3')
         except Exception as e:
            import logging
            log = logging.getLogger('root')
            log.info(f'processing parser of Project {project_name} failed with error:\ {e}')
            p = DefaultProjectArgparser()
            p.add_parser(subparser,project_name,project_path)

def create_experiment_parsers(parser,known_experiments,experiments_to_load):
   if not isinstance(experiments_to_load,(list,tuple)):
      experiments_to_load = tuple(known_projects.keys())
   subparser = parser.add_subparsers(help='Available xframe Experiments')      
   for exp_name in known_experiments:
      if exp_name in experiments_to_load:
         try:
            _argparser_ = importlib.import_module('xframe.experiments.'+exp_name+'._argparser_')
            exp_parser = subparser.add_parser(exp_name, help=_argparser_.help)
            #_argparser_.add_to_parser(exp_parser)               
         except Exception as e:
            import logging
            log = logging.getLogger('root')
            log.info(f'processing parser of Project {project_name} failed with error:\ {e}')
            project_parser = subparser.add_parser(project_name, help='No help available')

def create_argument_parser(defaults,xframe):
   p_name = defaults['project_name']
   p_opt = defaults['project_opt']
   e_name = defaults['exp_name']
   e_opt = defaults['exp_opt']
   xframe_help = f'xFrame {xframe.__version__}: A framwork for scientific computing targeting X-ray scattering experiments.'
   parser = argparse.ArgumentParser(prog='xframe', description=xframe_help)
   #parser.add_argument('ana_name', help='analysisWorker name', nargs = '?',default = a_name , type=str)
   parser.add_argument('-e','--experiment',choices=list(xframe.known_experiments.keys()), help='Available Experiments',default = e_name , type=str)
   parser.add_argument('-eset','--experiment_settings', help='experiment settings to be used',metavar='FILE_NAME',default = e_opt , type=str)
   parser.add_argument('-v','--verbose', help='Set loglevel to INFO.',action='store_true')
   parser.add_argument('-d','--debug', help='Set loglevel to DEBUG.',action='store_true')
   parser.add_argument('--version',action='store_true')
   parser.add_argument('--setup_home',metavar='HOME_PATH',nargs = '?',const='~/.xframe/',default=False, help='setup xframe home directory at HOME_PATH default is ~/.xframe')
   parser.add_argument('--print_home',action='store_true', help='show current home path')
   
   create_project_parsers(parser,xframe.known_projects,xframe.settings.general.load_projects,p_name)
   return parser

def start_routine_argparse():
   import xframe
   import logging

   defaults = {'project_name':False,'project_opt':False,'exp_name':False,'exp_opt':False}
   parser = create_argument_parser(defaults,xframe)
   args=parser.parse_args()
   #print(f'parsed args = {args}')
   if args.debug:
      logging.getLogger('root').setLevel(logging.DEBUG)
   elif args.verbose:
      logging.getLogger('root').setLevel(logging.INFO)   

      
   if args.setup_home:      
      from xframe.startup_routines import setup_home
      setup_home(args.setup_home)
   if args.print_home:
      xframe.lib.python.xprint(f'Current xframe home directory is: {xframe.settings.general.home_folder}')
   if args.version:
      xframe.lib.python.xprint(f'xFrame {xframe.__version__}')

   if not isinstance(args.project,bool):
      results = xframe.select_and_run(project = args.project, project_worker=args.worker,project_settings=args.settings,exp_name=args.experiment,exp_settings=args.experiment_settings,oneshot = True)
      xframe.Multiprocessing.update_free_cpus()
   else:
      logging.getLogger('root').info('No project name was given, terminating.')
   sys.exit()



def start_routine_click(click):
   import xframe
   import logging

   xprint = xframe.library.pythonLibrary.xprint
   db = xframe.database.default
   
   startup_dict = {'experiment_name':False,'exp_settings':False}
   known_experiments = xframe.known_experiments
   xframe_help = f'xFrame {xframe.__version__}: A framwork for scientific algorithms targeting X-ray scattering'
   
   @click.group(invoke_without_command=True,help=xframe_help)
   @click.option('--setup_home',is_flag = False,flag_value=xframe.settings.general.default_folder,is_eager=True,type=click.Path(exists=False,resolve_path=True), help='create home folder. If PATH not specified use ~/.xframe')
   @click.option('--print_home',is_flag = True,is_eager=True, help='print path of current home folder')
   @click.option('-e','--experiment',is_flag = False,flag_value=False,help='specify optional experiment module',type=click.Choice(list(known_experiments.keys()), case_sensitive=True))
   @click.option('-eopt','--experiment_options',is_flag = False,flag_value=False,help='settings name for experiment module',type=str)
   @click.option('-d','--debug', is_flag=True, help = 'set loglevel to DEBUG')
   @click.option('-v','--verbose', is_flag = True, help = 'set loglevel to INFO')
   @click.option('--version', is_flag = True,is_eager=True)
   @click.pass_context
   def run(ctx,*args,**kwargs):
      if kwargs['debug']:
         logging.getLogger('root').setLevel(logging.DEBUG)
         if ctx.invoked_subcommand is None:
            print(ctx.get_help())
      elif kwargs['verbose']:
         logging.getLogger('root').setLevel(logging.INFO)
         if ctx.invoked_subcommand is None:
            print(ctx.get_help())
      if isinstance(kwargs['setup_home'],str):
         xframe.startup_routines.setup_home(kwargs['setup_home'])
      if kwargs['print_home']:
         xprint(f"xframe's current home directory is: {xframe.settings.general.home_folder}")
      if kwargs['version']:
         xprint(f'xFrame {xframe.__version__}')
      if isinstance(kwargs['experiment'],str):
         startup_dict['experiment_name']=kwargs['experiment']
         xframe.select_experiment(exp_name=kwargs['experiment'],exp_settings=kwargs['experiment_options'])
         if ctx.invoked_subcommand is None:
            print(ctx.get_help())

      @ctx.call_on_close
      def close_workers():
         xframe.controller.control_worker.stop_working()
         del(xframe.controller.control_worker.gpu_manager)

   from xframe.startup_routines import DefaultProjectClick
   projects_to_load = xframe.settings.general.load_projects
   known_projects = xframe.known_projects
   if not isinstance(projects_to_load,(list,tuple)):
      projects_to_load = tuple(known_projects.keys())
   for project_name,project_path in known_projects.items():
      if project_name in projects_to_load:
         try:
            _argparser_ = importlib.import_module('xframe.projects.'+project_name+'._argparser_')
            p = getattr(_argparser_,'ProjectClick',DefaultProjectClick)()
            p.add_click(run,click,project_name,project_path)
         except Exception as e:
            import logging
            log = logging.getLogger('root')
            log.info(f'processing parser of Project {project_name} failed with error:\ {e}.Use default.')
            p = DefaultProjectClick()
            p.add_click(run,click,project_name,project_path)
   run()

   
def start_routine_cmd():
   import xframe
   try:
      click = xframe.interfaces.ClickDependency.click
      start_routine_click(click)
   except xframe.interfaces.DependencyImportError as e :
      log.info('Could not import click. Fallback to to argparse.')
      start_routine_argparse()
   
if __name__ == '__main__':
   variables = start_routine_cmd()



from xframe.interfaces import DependencyMeta
import abc
class SharedMemoryInterface(abc.ABC):
    #_dict structure is as follows
    # {mem_name:{'shape': <tuple(dim1,...)>,'dtype': <np.dtype object >}}
    _external_dependency_name_ = 'shared_array'
    @abc.abstractmethod
    def allocate(self,_dict):
        pass
    @abc.abstractmethod
    def delete(self,mem_names):
        pass
    @abc.abstractmethod
    def attach(self,mem_names):
        pass
    @abc.abstractmethod
    def detach(self,mem_names):
        pass    
    @abc.abstractmethod
    def __getitem__(self,index):
        pass
    @abc.abstractmethod
    def __len__(self):
        pass
    @abc.abstractmethod
    def __iter__(self):
        pass

class SharedMemoryDependency(SharedMemoryInterface,metaclass = DependencyMeta):
    pass
    
class ClFunctionInterface(abc.ABC):
    @abc.abstractmethod
    def assemble_buffers(self,kernel_data):
        pass
    @abc.abstractmethod
    def assemble_function(self,kernel_data):
        pass
class ClProcessInterface(abc.ABC):
    @abc.abstractmethod
    def assemble(self,kernel_data):
        pass
    
class OpenClInterface(abc.ABC):
    _external_dependency_name_ = 'opencl'
    ClProcess = ClProcessInterface
    ClFunction = ClFunctionInterface
    @abc.abstractmethod
    def create_context(self):
        pass
    @abc.abstractmethod
    def create_functions(self,kernel_data):
        pass
    @abc.abstractmethod
    def create_process_on_all_gpus(self,kernel_data):
        pass
    @abc.abstractmethod
    def get_number_of_gpus(self,kernel_data):
        pass


class OpenClDependency(OpenClInterface,metaclass=DependencyMeta):
    pass

class PsutilInterface(abc.ABC):
    _external_dependency_name_ = 'psutil'
    @abc.abstractmethod
    def get_free_memory(self):
        pass

class PsutilDependency(PsutilInterface,metaclass = DependencyMeta):
    pass

class SharedArrayDependency(metaclass = DependencyMeta):
    _external_dependency_name_ = 'shared_array'



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D_multi,plot1D
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi

ccd = db.load('ccd')
cc = ccd['cross_correlation']


def create_functions():
    cld = Multiprocessing.load_openCL_dict()
    cl = cld['cl']
    ctx = cld['context']
    queue = cld['queue']
    kernel = cl.Program(ctx, """
    __kernel void
    count_negative(__global double* I, 
    __global double* summands, 
    __global double* neg_counts, 
    long nN,long nq,long ntheta, long nphi)
    {
    
    long N = get_global_id(0); 
    long q = get_global_id(1); 
    long theta = get_global_id(2); 
    long phi = get_global_id(3); 
    
    // value stores the element that is 
    // computed by the thread
    double neg_count = 0;
    for (int phi = 0; phi < nphi; ++phi)
    {
    double sum = I[q*ntheta*nphi+theta*nphi+phi] + summands[N*nq+q]; 
    neg_count += (1.0-sum/fabs(sum)); // 2 if sum is negative, 0 else
    }
    // Write the matrix to device memory each 
    // thread writes one element
    neg_counts[N*nq*ntheta + q*ntheta + theta] = neg_count/2;
    }    
    __kernel void 
    floatSum(__global double* inVector, __global double* outVector, const int inVectorSize, __local double* resultScratch){
    int gid = get_global_id(0);
    int wid = get_local_id(0);
    int wsize = get_local_size(0);
    int grid = get_group_id(0);
    int grcount = get_num_groups(0);

    int i;
    int workAmount = inVectorSize/grcount;
    int startOffset = workAmount * grid + wid;
    int maxOffset = workAmount * (grid + 1);
    if(maxOffset > inVectorSize){
        maxOffset = inVectorSize;
    }
    resultScratch[wid] = 0.0;
    for(i=startOffset;i<maxOffset;i+=wsize){
            resultScratch[wid] += inVector[i];
    }
    barrier(CLK_LOCAL_MEM_FENCE);

    if(gid == 0){
    for(i=1;i<wsize;i++){
    resultScratch[0] += resultScratch[i];
    }
    outVector[grid] = resultScratch[0];
    }
    }


    __kernel void
    reduce(__global double* buffer,
    __global double* result,
    __const int length,
    __local double* scratch
    ) {
    int global_index = get_global_id(0);
    double accumulator = INFINITY;
    // Loop sequentially over chunks of input vector
    while (global_index < length) {
    double element = buffer[global_index];
    accumulator = (accumulator < element) ?
    accumulator : element;
    global_index += get_global_size(0);
    }
    // Perform parallel reduction
    int local_id = get_local_id(0);
    scratch[local_id]=accumulator;
    barrier(CLK_LOCAL_MEM_FENCE);
    
    for(int offset = get_local_size(0)/2;
    offset > 0;
    offset = offset / 2) {
    if (local_id<offset){
    double other = scratch[local_id+offset];
    double mine = scratch[local_id];
    scratch[local_id]= (mine<other) ? mine : other;
    }
    barrier(CLK_LOCAL_MEM_FENCE);
    }
    }
    
    """).build()
    sum1 = kernel.floatSum
    sum2 = kernel.reduce
    return locals()

locals().update(create_functions())

n=60000
a = np.ones(n,dtype = np.float64)

local_range = (np.int32(32),)
global_range = (np.int32(n),)

scratch=np.zeros(local_range,dtype = np.float64)
out = np.zeros(n,dtype = np.float64)

mf = cl.mem_flags
in_buff = cl.Buffer(ctx, mf.READ_WRITE,size=a.nbytes)
out_buff = cl.Buffer(ctx , mf.READ_WRITE, size=out.nbytes)
scratch_buff = cl.Buffer(ctx , mf.READ_WRITE, size=scratch.nbytes)


cl.enqueue_copy(queue,in_buff,np.ascontiguousarray(a))
sum2(queue,global_range,local_range,in_buff,out_buff,np.int32(n),cl.LocalMemory(a.nbytes))
cl.enqueue_copy(queue, out_buff,out)

#global_range_count = (nN,nq,ntheta)
#global_range_sum = (nN,)
#neg_counts = np.zeros(nN,dtype = float)
#I_mock = np.zeros((nq,ntheta,nphi),dtype = float)
#
#mf = cl.mem_flags
#summands_buff = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=np.ascontiguousarray(summands))
#I_buff = cl.Buffer(ctx , mf.READ_WRITE, size=I_mock.nbytes)
#out_buff = cl.Buffer(ctx , mf.READ_WRITE, size=neg_counts.nbytes)



import numpy as np
import numpy.ma as mp
import logging

from xframe.library import pythonLibrary as pyLib
from xframe.library.gridLibrary import NestedArray
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library import mathLibrary as mLib
from xframe.interfaces import DatabaseInterface
from xframe.interfaces import PresenterInterface
from xframe.presenters.matplotlibPresenter import heatPolar2D
pres=heatPolar2D()
log=logging.getLogger('root')

def getAnalysisRecipeFacotry(database,presenters):
    def getAnalysisOperatorDict(database,presenters):
        def getDataBaseOperatorDict(database):
            interfaceFunctionNames=pyLib.getFunctionNames(DatabaseInterface)
            allMethodsDict=pyLib.getMethods(database, _type='dict')
            dataDict={functionName+'@Database':allMethodsDict[functionName] for functionName in interfaceFunctionNames}
            return dataDict
        
        def getPresenterOperatorDict(presenters):
            interfaceFunctionNames=pyLib.getFunctionNames(PresenterInterface)
            presenterDict={}
            for presenter in presenters:
                Name=type(presenter).__name__
                methodsDict=pyLib.getMethods(presenter, _type='dict')
                tempDict={functionName+'@'+Name:methodsDict[functionName] for functionName in interfaceFunctionNames}
                presenterDict.update(tempDict)
            return presenterDict
    
        operatorDict={}
        analysisLibModule=__import__('analysisLibrary')  #fromlist is neccessary otherwise it only imports toplevel module analysis instead of analysis.analysusLibrary
        libraryDict=pyLib.getFunctions(analysisLibModule, _type='dict',recoursive=True)
#        print('this is the libraryDict: {} \n'.format(libraryDict))
#        print('current Module name = {}'.format(currentModule))
        operatorDict.update(libraryDict)
        mathDict=pyLib.getFunctions(mLib, _type='dict')
        operatorDict.update(mathDict)
        dataDict=getDataBaseOperatorDict(database)
        operatorDict.update(dataDict)        
        presenterDict=getPresenterOperatorDict(presenters)
        operatorDict.update(presenterDict)        
        return operatorDict

    operatorDict= getAnalysisOperatorDict(database,presenters)
    recipeFactory=pyLib.RecipeFactory(operatorDict)
    return recipeFactory


def get_analysis_process_factory(database,presenters):
    def getAnalysisOperatorDict(database,presenters):
        def getDataBaseOperatorDict(database):
            interfaceFunctionNames=pyLib.getFunctionNames(DatabaseInterface)
            allMethodsDict=pyLib.getMethods(database, _type='dict')
            dataDict={functionName+'@Database':allMethodsDict[functionName] for functionName in interfaceFunctionNames}
            return dataDict
        
        def getPresenterOperatorDict(presenters):
            interfaceFunctionNames=pyLib.getFunctionNames(PresenterInterface)
            presenterDict={}
            for presenter in presenters:
                Name=type(presenter).__name__
                methodsDict=pyLib.getMethods(presenter, _type='dict')
                tempDict={functionName+'@'+Name:methodsDict[functionName] for functionName in interfaceFunctionNames}
                presenterDict.update(tempDict)
            return presenterDict
    
        operatorDict={}
        analysisLibModule=__import__('analysisLibrary')  #fromlist is neccessary otherwise it only imports toplevel module analysis instead of analysis.analysusLibrary
        libraryDict=pyLib.getFunctions(analysisLibModule, _type='dict',recoursive=True)
#        print('this is the libraryDict: {} \n'.format(libraryDict))
#        print('current Module name = {}'.format(currentModule))
        operatorDict.update(libraryDict)
        mathDict=pyLib.getFunctions(mLib, _type='dict')
        operatorDict.update(mathDict)
        dataDict=getDataBaseOperatorDict(database)
        operatorDict.update(dataDict)        
        presenterDict=getPresenterOperatorDict(presenters)
        operatorDict.update(presenterDict)        
        return operatorDict

    operatorDict= getAnalysisOperatorDict(database,presenters)
    recipeFactory=pyLib.RecipeFactory(operatorDict)
    return recipeFactory



def returnGrid_deleteFirstDimension(oldGrid):
    def deleteFirstCoordinate(point):
        return np.delete(point,0)
    reducedGridArray=np.array(tuple(map(deleteFirstCoordinate,oldGrid[0]))).reshape(oldGrid.array.shape[1:]+(oldGrid.dimension-1,))
    reducedGridArray=pyLib.getArrayOfArray(reducedGridArray)
    reducedGrid=Grid(reducedGridArray,oldGrid.gridType)
    return reducedGrid


def generate_save_to_dict(dictionary):
    def save_to_dict(key,mode,data):
        if mode == 'append':
            if key in dictionary:
                dictionary[key].append(data)
            else:
                dictionary[key]=[data]
        elif mode=='overwrite':            
            dictionary[key]=data
        else:
            dictionary[key]=data
        return data
    return save_to_dict
def generate_load_from_dict(dictionary):
    def load_from_dict(key):
        value=dictionary[key]
        return value
    return load_from_dict

        
def copy(data):
    return data.copy()

def square_grid(data):
    return (data*data.conjugate()).real

def abs_value(data):
    return np.abs(data)


def generate_calc_center(real_grid):
    dim = real_grid.n_shape[0]
    cart_grid = spherical_to_cartesian(real_grid)    
    def calc_center(density):
        density_sum=np.sum((density.real*real_grid[...,0]).flatten())
        if density_sum==0:
            density_sum=1
        center=np.sum((cart_grid*(real_grid[...,0]*density.real)[...,None]).reshape(-1,dim),axis=0)/density_sum
        log.info('center={}'.format(center))
        center=cartesian_to_spherical(center)
        #polygon=mLib.SampleShapeFunctions.get_disk_function(0.05,amplitude_function=lambda points: np.full(points.shape[:-1],0.01),coordSys=1,center=center)
        #temp_density=np.abs(density)+polygon(real_grid)
        #pres.present(temp_density)
        log.info('center={}'.format(center))
        return center        
    return calc_center
    
    

class Selection:
    def __init__(self,name,selection_dict):
        self.name = name
        self.data_range= selection_dict.get('range',slice(None))
        self.range_type=self.data_range.__class__
        if self.range_type != slice:
            self.data_range = np.asarray(self.data_range)
            self.range_type=self.data_range.__class__
        self.mode = selection_dict.get('mode','relative')
    def mask(self,input_data):
        if self.range_type == slice:
            mask = self.slice_to_mask(input_data)
        else:
            mask = self.numpy_to_mask(input_data)
        return mask
    def slice_to_mask(self,input_data):
        u_input = np.unique(input_data)
        _slice = self.data_range
        if self.mode == 'relative':
            selection_array = u_input[_slice]
        else:
            _min,_max = u_input.min(),u_input.max()
            start = _slice.start
            stop = _slice.stop
            step = _slice.step
            if start == None:
                start = _min
            if stop == None:
                stop == _max
            if step == None:
                step = int((stop -start)/abs(stop-start))
            selection_array = np.array(range(start,stop,step))
        mask = self._array_to_mask(input_data,selection_array)
        return mask
    
    def numpy_to_mask(self,input_data):
        u_input = np.unique(input_data)
        _array = self.data_range
        if self.mode == 'relative':
            try:
                selection_array = u_input[_array]
            except IndexError as e:
                log.error('Selection of {}: Specified relative indices {} are out of range for the {} unique {}. Continue with empty selection.'.format(self.name,_array,len(u_input),self.name))
                selection_array = np.array([])
        else:
            selection_array = _array
        mask = self._array_to_mask(input_data,selection_array)
        return mask
    
    def _array_to_mask(self,input_data,array):
        return np.isin(input_data,array)

        



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import importlib
from scipy.stats import unitary_group
from scipy.stats import special_ortho_group
from xframe import log
#from testing.testLibrary import defaultTest_array
from xframe.library import pythonLibrary as pyLib
from xframe.library import mathLibrary as mLib
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import uniformGrid_func
from xframe.library.gridLibrary import GridFactory
from xframe.library.gridLibrary import ReGrider
from xframe.library.gridLibrary import SampledFunction
from xframe.library.gridLibrary import NestedArray
from xframe.database.general import load_settings
from xframe.externalLibraries.shtns_plugin import sh as shtns
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid_new as get_grid
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft 
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_spherical_zernike_ht 
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection
from xframe.control.communicators import mock_comm_module_single_thread
from xframe.plugins.MTIP.analysisLibrary.misk import generate_calc_center
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import generate_shift_by_operator

from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import calc_spherical_zernike_weights
from xframe.plugins.MTIP.database import Analysis_DB
import vtk
from vtk.util import numpy_support as vtn
from xframe.control.communicators import mock_comm_module_single_thread

log=log.setup_custom_logger('root','INFO')
comm_module = mock_comm_module_single_thread()


def initialize(l_max=99, n_points=200,ft_mode='Zernike_3D',pi_in_q = False):
    opt,_ = load_settings('plugins/MTIP','3d_model_0')    
    db = Analysis_DB(**opt['IO'])
    mLib.shtns = shtns

    harm_opt={'dimensions':3,
              'indices':'lm',
              'max_order':l_max}#19}

    ht = HarmonicTransform('complex',harm_opt)

    if ft_mode == 'Zernike_3D':
        grid_type='Zernike'
        Q=0.05141790928385248
        if pi_in_q:            
            Q*=2*np.pi
            
    elif ft_mode == 'PNAS_3D':
        grid_type = 'PNAS'
        Q=n_points/2
        
    opt = {
        **{
            'dimensions':3,
            'type':grid_type,
            'reciprocal_cut_off':Q,
            'n_radial_points':n_points,
            'pi_in_q':pi_in_q
        },**ht.grid_param
    }

    opt2 = {
        'dimensions':2,
        'type':'Zernike',
        'reciprocal_cut_off':0.05141790928385248,
        'n_radial_points':200,
        'n_angular_points':1600,
        'pi_in_q':pi_in_q
    }
    
    grid_pair =  get_grid(opt)
    grid = grid_pair.realGrid
    grid2 = get_grid(opt2).realGrid
    r_grid = grid_pair.reciprocalGrid
    max_r = grid[:,0,0,0].max()
    n_radial_points = grid.shape[0]
    max_order = harm_opt['max_order']
    n_orders = max_order+1
    name_postfix='N'+str(n_radial_points)+'mO'+str(max_order)+'nO'+str(n_orders)
    #name_postfix='N'+str(n_radial_points)+'mO'+str(29)+'nO'+str(30)

    if ft_mode == 'Zernike_3D':
        pos_orders=np.arange(n_orders)
        expansion_limit = 1000
        opt={'expansion_limit':expansion_limit,'max_radius':max_r,'n_radial_points':n_radial_points,'pi_in_q':pi_in_q}
        weights=calc_spherical_zernike_weights(pos_orders,opt)
        weights_dict={'weights':weights,'maxR':max_r,'posHarmOrders':pos_orders}        
        ft,ift = generate_zernike_spherical_ft(max_r,weights_dict,ht,'complex',pi_in_q)
        zht,izht = generate_spherical_zernike_ht(weights_dict['weights'],weights_dict['posHarmOrders'],max_r,pi_in_q=pi_in_q)
    elif ft_mode == 'PNAS_3D':
        log.info('henerate PNAS versions of transforms')
        weights_dict=db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':'PNAS_3D'})
        ft,ift = generate_Donatelli_spherical_ft(weights_dict,ht,'complex') 
        zht,izht = generate_ht_spherical_SinCos(weights_dict)
        
    cube_func = mLib.SampleShapeFunctions.get_disk_function(500,norm='inf',random_orientation = True)
        
    return db,ht,ft,ift,grid_pair,grid2,zht,izht

os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe/xframe/')
db,ht,ft,ift,grid_pair,grid2,zht,izht = initialize(n_points = 128,pi_in_q = True)

f_ball = mLib.SampleShapeFunctions.get_disk_function(10,center=(40,np.pi/2,np.pi/2))
f_ball2 = mLib.SampleShapeFunctions.get_disk_function(10,center=(40,np.pi/2,3/2*np.pi))
ball = f_ball(grid_pair.realGrid)
balls = ball + 2*f_ball2(grid_pair.realGrid)
calc_center = generate_calc_center(grid_pair.realGrid)
shift = generate_shift_by_operator(grid_pair.reciprocalGrid,opposite_direction = True)

center = calc_center(balls) 
a_balls = ft(balls)
a_balls = shift(a_balls,center)
balls2 = ift(a_balls)
center2 = calc_center(balls2) 



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_covid')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)


def projection_matrix_new():
    try:
        ccd= db.load('ccd')
        ReciprocalProjection.pre_init_from_ccd(ccd)
        p2 = ReciprocalProjection.data_projection_matrices

        
    except Exception as e:
        traceback.print_exc()
    return locals()

locals().update(projection_matrix_new())



import abc


class PresenterInterface(abc.ABC):
    @abc.abstractmethod
    def present(self):
        pass
    @abc.abstractmethod
    def present2(self):
        pass


class DatabaseInterfaceAnalysis(abc.ABC):
    @abc.abstractmethod
    def load(self,analysisType,restultId):
        pass
    @abc.abstractmethod
    def save(self,analysisType,restultId):
        pass



import h5py as h5
import numpy as np
import os
from os.path import join as pjoin
from os import path as op
import time
import logging
import traceback
import re
import glob

from xframe.interfaces import DatabaseInterface,ExperimentWorkerInterface,ProjectWorkerInterface
from xframe.library.mathLibrary import plane3D
from xframe.library.mathLibrary import spherical_to_cartesian
#from xframe.projectRecipes import analysisLibrary as aLib
from xframe.library.gridLibrary import GridFactory
from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import double_first_dimension
from xframe.library.pythonLibrary import getArrayOfArray
from xframe.library.pythonLibrary import option_switch
from xframe.library.pythonLibrary import xprint
from xframe.library.physicsLibrary import ewald_sphere_theta
from xframe.library.physicsLibrary import ewald_sphere_theta_pi
from xframe.database.database import DefaultDB
from .projectLibrary.classes import FXS_Data
from .projectLibrary.classes import FTGridPair
from .projectLibrary.classes import SampledFunction
from .projectLibrary.fxs_invariant_tools import ccd_associated_legendre_matrices_single_m
from .projectLibrary.fxs_invariant_tools import deg2_invariant_to_cn_3d
from .projectLibrary.fxs_invariant_tools import harmonic_coeff_to_deg2_invariants
from .projectLibrary.misk import _get_reciprocity_coefficient
from .projectLibrary.harmonic_transforms import HarmonicTransform

from xframe.presenters.matplotlibPresenter import heat2D_multi
from xframe.presenters.matplotlibPresenter import plot1D
from xframe.presenters.openCVPresenter import Polar2D
from xframe import settings
import xframe


log=logging.getLogger('root')
        
class ProjectDB(DefaultDB,DatabaseInterface):
    def __init__(self,**folders_files):
        super().__init__(**folders_files)


    def get_time_string(self):
        time_struct=time.gmtime()
        time_str=str(time_struct[2])+'_'+str(time_struct[1])+'_'+str(time_struct[0])
        return time_str
    def get_reconstruction_path(self):
        time_str=self.get_time_string()
        path=self.folders[self.files['reconstructions']['folder']]
        run= self.get_latest_run('reconstructions',path_modifiers={'time':time_str}) + 1
        run_path=path.format(time=time_str,run=run)
        return run_path
    
    def get_latest_run(self,folder_name,path_modifiers = {}):
        '''
        Searches subfolders for increasing run number.
        Stops if either the folder does not exist or does not contain a file ending with .h5 
        '''
        if folder_name in self.files:
            path=self.folders[self.files[folder_name]['folder']]
        else:
            path = folder_name
        
        run=0
        path_modifiers['run']=run
        run_path=path.format(**path_modifiers)
        log.info(f'run_path={run_path}')
        while os.path.exists(run_path):
            #print(run_path)

            h5_found = False
            for file in os.listdir(run_path):
                # check the files which are end with specific extension
                if file.endswith(".h5"):
                    h5_found=True

            if h5_found:
                run+=1
                path_modifiers['run']=run
                run_path=path.format(**path_modifiers)
            else:
                break
        return run-1

    def save_average_results(self,name,data,**kwargs):
        grid_pair = data['internal_grid']
        real_grid = grid_pair["real_grid"]
        reciprocal_grid = grid_pair["reciprocal_grid"].copy()
        dimension = int(real_grid[:].shape[-1])
        
        
        options = kwargs
        log.info(options)
        time_struct=time.gmtime()
        time_str=str(time_struct[2])+'_'+str(time_struct[1])+'_'+str(time_struct[0])
        path=self.folders[self.files['average_results']['folder']]
        path_modifiers={'time':time_str,'structure_name':settings.project.structure_name,'dimensions':dimension}
        run= self.get_latest_run('average_results',path_modifiers=path_modifiers) + 1
        path_modifiers['run']=run
        run_path=path.format(**path_modifiers)
        log.info('run_path = {}'.format(run_path))
        path = self.get_path('average_results', path_modifiers=path_modifiers)
        self.save(path,data)

        if options.get('save_settings',True):
            path = os.path.dirname(self.get_path('average_results', path_modifiers=path_modifiers))
            self._save_settings(path)

        grid_pair = data['internal_grid']
        real_grid = grid_pair["real_grid"]
        reciprocal_grid = grid_pair["reciprocal_grid"].copy()
        dimension = int(real_grid[:].shape[-1])
        qs = reciprocal_grid[:].__getitem__((slice(None),)+(int(0),)*dimension)
        #reciprocity_coefficient = data['reciprocity_coefficient']
        #reciprocal_grid[...,0]*= (np.pi/reciprocity_coefficient) 
        vtk_saver = self.get_db('file://vtk').save

        if options['generate_average_vtk']:
            try:
                real_density = data['average']['real_density']
                normalized_real_density = data['average']['normalized_real_density']
                reciprocal_density = data['average']['reciprocal_density']

                if dimension == 2:
                    grid_type = 'polar'
                elif dimension == 3:
                    grid_type = 'spherical'
                real_vtk_path=self.get_path('real_vtk',path_modifiers={**path_modifiers,'reconstruction':'average'})
                self.save(real_vtk_path,[real_density.real,normalized_real_density.real],grid = real_grid,dset_names=['density','normalized_density'],grid_type=grid_type)
                #vtk_saver([real_density],real_grid,real_vtk_path,dset_names = ['density'],grid_type='spherical')

                reciprocal_vtk_path=self.get_path('reciprocal_vtk',path_modifiers={**path_modifiers,'reconstruction':'average'})
                self.save(reciprocal_vtk_path,[np.abs(reciprocal_density).real**2],grid = reciprocal_grid,dset_names=['amplitude'],grid_type=grid_type)
                #vtk_saver([reciprocal_density],reciprocal_grid,reciprocal_vtk_path, dset_names = ['amplitude'],grid_type='spherical')
            except Exception as e:
                traceback.print_exc()
                log.error('Failed to generate aligned vtk.')

                
        if options['generate_aligned_vtk']:
            try:
                for key,result in data['aligned'].items():
                    real_density = result['real_density']
                    #log.info('averaged shape = {}'.format(real_density.shape))
                    #log.info('averaged grid shape = {}'.format(real_grid.shape))
                    reciprocal_density = result['reciprocal_density']
                    if dimension == 2:
                        grid_type = 'polar'
                    elif dimension == 3:
                        grid_type = 'spherical'                        
                    #vtk_saver([real_density],real_grid,real_vtk_path, dset_names = ['density'],grid_type='spherical')
                    real_vtk_path=self.get_path('real_vtk',path_modifiers={'time':time_str,'run':run,'reconstruction':'aligned_'+key})
                    #log.info(real_vtk_path)
                    self.save(real_vtk_path,[real_density.real],grid = real_grid,dset_names=['density'],grid_type=grid_type)
                    reciprocal_vtk_path=self.get_path('reciprocal_vtk',path_modifiers={**path_modifiers,'reconstruction':'aligned_'+key})
                    self.save(reciprocal_vtk_path,[np.abs(reciprocal_density).real**2],grid = reciprocal_grid,dset_names=['amplitude'],grid_type=grid_type)
            except Exception as e:
                traceback.print_exc()
                log.error('Failed to generate aligned vtk.')                        

        log.info('input vtk is set to = {}'.format(options['generate_input_vtk']))
        if options['generate_input_vtk']:
            try:
                log.info('len input data = {}'.format(len(data['input'].items())))
                for key,result in data['input'].items():
                    #log.info("key = {}".format(key))
                    real_density = result['real_density']
                    support_mask = result['support_mask']
                    reciprocal_density = result['reciprocal_density']

                    if dimension == 2:
                        grid_type = 'polar'
                    elif dimension == 3:
                        grid_type = 'spherical'
                    
                    real_vtk_path=self.get_path('real_vtk',path_modifiers={**path_modifiers,'reconstruction':'input_'+key})
                    self.save(real_vtk_path,[real_density.real,support_mask],grid = real_grid,dset_names=['density','mask'],grid_type=grid_type)
                    #vtk_saver([real_density,support_mask],real_grid,real_vtk_path, dset_names = ['density','mask'],grid_type='spherical')

                    reciprocal_vtk_path=self.get_path('reciprocal_vtk',path_modifiers={**path_modifiers,'reconstruction':'input_'+key})
                    #vtk_saver([reciprocal_density],reciprocal_grid,reciprocal_vtk_path , dset_names = ['amplitude'],grid_type='spherical')
                    self.save(reciprocal_vtk_path,[np.abs(reciprocal_density).real**2],grid = reciprocal_grid,dset_names=['amplitude'],grid_type=grid_type)
            except Exception as e:
                traceback.print_exc()
                log.error('Failed to generate input vtk.')
                
        if options['generate_rotation_metric_vtk']:
            try:
                if dimension == 3:
                    for key,metrics in data['rotation_metric'].items():
                        for n,metric in enumerate(metrics):
                            vtk_path = self.get_path('rotation_metric_vtk',path_modifiers={**path_modifiers,'reconstruction':key,'iteration':str(n+1)})
                            self.save(vtk_path,[metric],grid = data['so3_grid'],dset_names=['rotation_metric'],grid_type='cartesian')
                            #vtk_saver([metric],data['so3_grid'],vtk_path,dset_names = ['rotation_metric'],grid_type='cartesian')
            except Exception as e:
                traceback.print_exc()
                log.error('Failed to generate rotation metric vtk.')
        if options['plot_resolution_metrics']:
            try:
                for key,metrics in data['resolution_metrics'].items():
                    if key == "PRTF":
                        layout = {'title':'PRTF','x_label':'q [\AA$^{-1}$]','y_label':'PRTF [arb.]'}                        
                        fig = plot1D.get_fig(np.abs(metrics),grid = qs,ylim=[0.0,1.1],layout = layout,labels=['PRTF'])
                        fig_path = run_path + 'PRTF.matplotlib'
                        self.save(fig_path,fig)
            except Exception as e:
                traceback.print_exc()
                log.error('Failed to generate resolution metric plots.')

    def _save_settings(self,path,project_name='settings',exp_name='exp_settings'):
        project_settings, experiment_settings = settings.get_settings_to_save()
        if isinstance(xframe.project_worker,ProjectWorkerInterface):
            self.save(pjoin(path,project_name+'.yaml'),project_settings,convert=True)
        if isinstance(xframe.experiment,ExperimentWorkerInterface):
            self.save(pjoin(path,exp_name+'.yaml'),experiment_settings,convert=True)
        #log.info(f'settings raw = \n {settings.raw_analysis} \n')
    
    def save_reconstructions(self,name,data,**kwargs):
        options = self.files['reconstructions']['options']
        time_str=self.get_time_string()
        path=self.folders[self.files['reconstructions']['folder']]
        path_modifiers={'time':time_str,'structure_name':settings.project.structure_name,'dimensions':settings.project.dimensions}     
        run= self.get_latest_run('reconstructions',path_modifiers=path_modifiers) + 1
        path_modifiers['run']=run
        run_path=path.format(**path_modifiers)

        internal_grids = data['configuration']['internal_grid']
        real_grid  = internal_grids['real_grid'].copy()
        reciprocal_grid  = internal_grids['reciprocal_grid'].copy()
        q_radial_points = reciprocal_grid.__getitem__((slice(None),)+(0,)*settings.project.dimensions)
        #log.info(reciprocal_grid[:,0,0])
        #reciprocity_coefficient = data['configuration']['reciprocity_coefficient']
        #log.info('reciprocity_coefficient in saving = {}'.format(reciprocity_coefficient))
        #reciprocal_grid[...,0]*=(np.pi/reciprocity_coefficient)
        
        log.info('run_path = {}'.format(run_path))
        if 'projection_matrices' in data:
            p_matrices = data['projection_matrices']

        ####### Data & Settings #######
        self.save('reconstructions',data,skip_custom_methods=True,path_modifiers=path_modifiers)
        self._save_settings(run_path)

        

        ####### Error Metrics #######
        try:
            if options.get('plot_error_metrics',False):
                errors={}
                for key,value in data['reconstruction_results'].items():
                    errors[key] = value['error_dict']['main'][2:]
                error_fig = plot1D.get_fig(np.array(tuple(errors.values()))[::-1],labels = tuple(errors.keys())[::-1],y_scale='log',layout={'title':'Mean of Relative projection differences in L2 norm.','x_label':'loop step','y_label':'$|| rho-rho_{projected}||^2_{l_2}/ || rho ||^2_{l_2}$'})
                error_path = run_path + 'error_metrics.matplotlib'
                self.save(error_path,error_fig)
        except Exception as e:
            log.warning(f'Faild to save error metrics with error:\n {e}')
            log.info(traceback.format_exc())

        try:
            plot_invariant_error,err_orders = option_switch(options.get('plot_invariant_error',False),[0,2,4])
            if plot_invariant_error:
                r = data['reconstruction_results']
                log.info('contents are as follows = {}'.format([d['error_dict'].keys() for d in r.values()]))
                errors = np.moveaxis(np.array([d['error_dict']['reciprocal']['deg2_invariant_l2_diff'] for d in r.values()]),2,0)
                for order in err_orders:
                    if order in settings.project.projections.reciprocal.used_order_ids:
                        fig = plot1D.get_fig(errors[order,::-1,2:],y_scale = 'log',layout={'title':'Relative B{} errors.'.format(order),'x_label':'loop step','y_label':'$\sum_{q_1,q_2}(B_l-B_l^{data})^2/(B_l^{data})^2$'})
                        error_path = run_path + 'B{}_errors.matplotlib'.format(order)
                        self.save(error_path,fig)
                mean_error = np.mean(errors[::2],axis = 0)
                fig = plot1D.get_fig(mean_error[::-1,2:],y_scale = 'log',layout={'title':'Relative Bl errors mean.','x_label':'loop step','y_label':'$\sum_{q_1,q_2}(B_l-B_l^{data})^2/(B_l^{data})^2$'})
                error_path = run_path + 'Bl_mean_error.matplotlib'.format(order)
                self.save(error_path,fig)
        except Exception as e:
            log.warning(f"Failed to save invariant error plots with error {e}")
            log.info(traceback.format_exc())

        ####### VTK Files & Pics #######
        try:  
            generate_vtk_files,vtk_slice = option_switch(options['generate_vtk_files'],slice(None))
            if generate_vtk_files:                
                log.info('start constructing/saving vtk files')
                n_reconstructions = len(data['reconstruction_results'])
                ids = np.array(tuple(data['reconstruction_results'].keys()))
                good_ids= ids[vtk_slice]
                worst_id = [ids[-1]]
                if len(ids)==1:
                    worst_id=[]
                    ids_to_plot = np.asarray(good_ids)
                else:
                    ids_to_plot = np.unique(np.concatenate((good_ids,worst_id)))
                
                for id in ids_to_plot:
                    result = data['reconstruction_results'][id]
                    real_density = result['real_density'].real
                    real_mask = result['support_mask']
                    initial_density = result['initial_density'].real
                    initial_support = result['initial_support']
                    last_real_density = result['last_real_density'].real
                    last_real_mask = result['last_support_mask']
                    reciprocal_intensity = np.abs(result['reciprocal_density']).real
                    last_reciprocal_intensity = np.abs(result['last_reciprocal_density']).real

                    _id_str = str(id)
                    if id in worst_id:
                        _id_str = f'{id}_worst_error'
                    vtk_path_modifiers = {**path_modifiers,**{'reconstruction':_id_str}}
                    real_vtk_path=self.get_path('real_vtk',path_modifiers=vtk_path_modifiers)

                    if settings.project["dimensions"] == 3:

                        self.save(real_vtk_path,[real_density,real_mask,last_real_density,last_real_mask,initial_density,initial_support],grid =real_grid,grid_type='spherical',skip_custom_methods=True,dset_names=['best_density','best_mask','last_density','last_mask','initial_density','initial_support'])
                        #save_vtk([real_density,real_mask],real_grid,real_vtk_path,grid_type='spherical')
                         
                        reciprocal_vtk_path=self.get_path('reciprocal_vtk',path_modifiers=vtk_path_modifiers)
                        self.save(reciprocal_vtk_path,[reciprocal_intensity,last_reciprocal_intensity],grid = reciprocal_grid,grid_type='spherical',skip_custom_methods=True,dset_names=['best_intensity','last_intensity'])
                    elif settings.project["dimensions"] == 2:
                        self.save(real_vtk_path,[real_density,real_mask,last_real_density,last_real_mask,initial_density,initial_support],grid =real_grid,grid_type='polar',skip_custom_methods=True,dset_names=['best_density','best_mask','last_density','last_mask','initial_density','initial_support'])
                        #save_vtk([real_density,real_mask],real_grid,real_vtk_path,grid_type='spherical')
                         
                        reciprocal_vtk_path=self.get_path('reciprocal_vtk',path_modifiers=vtk_path_modifiers)
                        self.save(reciprocal_vtk_path,[reciprocal_intensity,last_reciprocal_intensity],grid = reciprocal_grid,grid_type='polar',skip_custom_methods=True,dset_names=['best_intensity','last_intensity'])
                        #log.info(reciprocal_grid[:,0,0])
                         
        except Exception as e:
            log.warning(f"Failed to save vtk_plots with error \n{e}")
            log.info(traceback.format_exc())
            
        try:
            generate_2d_images,img_slice = option_switch(options.get('generate_2d_images',False),slice(None))
            if generate_2d_images:                
                log.info('start constructing & saving polar images')
                for id,result in tuple(data['reconstruction_results'].items())[img_slice]:
                    self._save_polar_densities(id,result,path_modifiers)
        except Exception as e:
            log.warning(f"Failed to save 2D polar images with error\n {e}")
            log.info(traceback.format_exc())
            
        ####### Deg 2 invariants & Rest #######
        try:
            if options.get('plot_last_fqc_error',False):
                log.info('\n \n \n  trying to plot errors ')
                r = data['reconstruction_results']
                errors = np.array([d['error_dict']['reciprocal']['fqc_error'] for d in r.values()])
                fig = plot1D.get_fig(1-errors[:,-1,:],y_scale = 'lin',layout={'title':'FQC','x_label':'q','y_label':'FQC'})
                error_path = run_path + 'last_fqc.matplotlib'
                self.save(error_path,fig)
        except Exception as e:
            log.warning(f"Failed to save fqc error plots with error\n {e}")
            log.info(traceback.format_exc())
                                
        try:
            plot_reconstructed_deg2_invariants,deg2_slice = option_switch(options.get('plot_reconstructed_deg2_invariants',False),slice(None))
            if plot_reconstructed_deg2_invariants:
                for key,value in tuple(data['reconstruction_results'].items())[deg2_slice]:
                    Bls = value['last_deg2_invariant']
                    log.info("last deg2 invariant shape = {}".format(Bls.shape))
                    self._save_first_invariants(Bls,q_radial_points,run_path,options,name="{}_out_".format(key))


                    #thetas = ewald_sphere_theta_pi(data['configuration']['xray_wavelength'],q_radial_points)
                    #PlmPlm = np.moveaxis(np.squeeze(ccd_associated_legendre_matrices_single_m(thetas,len(p_matrices)-1,0)),-1,0)
                    #PlmPlm[np.isnan(PlmPlm)]=0
                    #PlmPlm[0]=0
                    #C0 = np.abs(np.sum(Bls*PlmPlm,axis = 0))
                    #self._save_C0(C0,q_radial_points,run_path,options,name="{}_out_".format(key))
        except Exception as e:
            log.warning(f"Failed to save invariant plots with error \n {e}")
            log.info(traceback.format_exc())
        try:
            if options.get('plot_first_used_invariants',False):
                if settings.project["dimensions"] == 3:
                    Bls = np.array([m @ m.T.conj() for m in p_matrices])
                elif settings.project["dimensions"] == 2:
                    Bls = np.array([m[:,None] * m.T.conj()[None,:] for m in p_matrices])
                #C0 = np.abs(Bls[0]-value['last_deg2_invariant'][0])**2/np.abs(Bls[0])**2
                #log.info('B[0] errors mean = {} = {}'.format(np.mean(C0),C0))
                #self._save_C0(C0,q_radial_points,run_path,options,name="Bl0_")
                self._save_first_invariants(Bls,q_radial_points,run_path,options,name="first_")               
        except Exception as e:
            log.warning(f"Failed to save initial invariant plots with error {e}")
            log.info(traceback.print_exc())
      
            
    def load_reconstruction_settings(self,name,**kwargs):
        path = self.get_path(name,path_modifiers=kwargs['path_modifiers'])        
        settings, raw_settings = self.load('settings',direct_path=path)
        return settings
    
    def _save_first_invariants(self,bls,radial_points,base_path,options,name='',mask = True):
        bls = bls.copy()
        bls[~mask]=0
        try:
            assert not np.isnan(bls).any(),'Nan values detected during plotting setting them to -1.'
        except:
            bls[np.isnan(bls)]=-1
            
        if isinstance(mask,np.ndarray):
            max_value = max(np.max(np.abs(bls[mask].flatten())),1e-100)
            median_value = np.median(np.abs(bls[::2][mask[::2]].flatten()))
        else:
            max_value = max(np.max(np.abs(bls.flatten())),1e-100)
            median_value = np.median(np.abs(bls[::2].flatten()))
        #log.info(f'median {median_value} max = {max_value}')
        if (median_value>0) and (median_value!=np.nan) and (max_value<np.inf):
            orders = int(np.log10(max_value/median_value)*0.8)
            vmin,vmax = options.get('plot_range',[median_value*10**(-orders//2),median_value*10**orders])
        elif (max_value==np.inf) and (median_value>0) and (median_value!=np.nan):            
            orders = 12
            vmin,vmax = options.get('plot_range',[median_value*10**(-orders//2),median_value*10**orders])
        else:            
            orders = 12
            vmin,vmax = options.get('plot_range',[max_value*10**(-orders//2),max_value])
            #log.info(f'vmin vmax = {vmin},{vmax}')
        #log.info(f'vmin vmax = {vmin}, {vmax} median = {median_value} max = {max_value}')
        
        grid = GridFactory.construct_grid('uniform',[radial_points,radial_points])
        order_ids = [0,2,4,6,8]

        shape = [3,len(order_ids)]
        layouts = []
        plot_data = []
        #log.info('bls shape = {} grid shape = {}'.format(bls.shape,grid.shape))
        for i in range(shape[0]):
            layout_part = []
            plot_data_part = []
            orders = np.arange(10*i,10*i+len(order_ids)*2,2)
            for o in orders:
                layout = {'title':'$B_{'+'{}'.format(o)+'}$',
                          'x_label':'$q_1$',
                          'y_label':'$q_2$'
                          }
                layout_part.append(layout)
                try:
                    plot_data_part.append(np.abs(bls[o]))
                except IndexError:
                    plot_data_part.append(np.zeros_like(bls[0],dtype = float))
                    log.info('IndexError')
                    
            layouts.append(layout_part)
            plot_data.append(plot_data_part)
                                                    
        fig_bl_masks = heat2D_multi.get_fig(plot_data,scale = 'log',layout = layouts,grid =grid[:],shape = shape,size = (30,shape[0]*5),vmin= vmin, vmax = vmax,cmap='plasma')
        bl_path = base_path +name+'Bl.matplotlib'
        self.save(bl_path,fig_bl_masks,dpi = 300)
    def _save_C0(self,c0,radial_points,base_path,options,name=''):
        max_value = np.abs(c0).max()
        vmin,vmax = options.get('plot_range_C1',[max_value*1e-12,max_value])
        if [vmin,vmax] == [None,None]:
            vmin,vmax = [max_value*1e-12,max_value]
    
        grid = GridFactory.construct_grid('uniform',[radial_points,radial_points])
        shape = [1,1]
        layout = {'title':'$C_0(q_1,q_2)$',
                  'x_label':'$q_1$',
                  'y_label':'$q_2$'
                  }
                                                    
        fig_bl_masks = heat2D_multi.get_fig(c0,scale = 'log',layout = [[layout]],grid =grid[:],size = (10,10),vmin= vmin, vmax = vmax,cmap='plasma')
        bl_path = base_path +name+'C0.matplotlib'
        self.save(bl_path,fig_bl_masks,dpi = 300)

    def _save_polar_densities(self,id,result,path_modifiers):
        #density_path=self.get_path('result_image_density',path_modifiers={'time':time_str,'run':run,'reconstruction':id})
        #intensity_path=self.get_path('result_image_density',path_modifiers={'time':time_str,'run':run,'reconstruction':id})
        pic_path=self.get_path('result_image',path_modifiers={**path_modifiers,'reconstruction':id})
        last_real_density = result['last_real_density'].real
        last_real_mask = result['last_support_mask']
        last_reciprocal_intensity = np.abs(result['last_reciprocal_density']).real
        dpic = Polar2D.get_fig(last_real_density,transparent_backgound=True,colormap='inferno',print_colorscale=True,n_pixels=512)
        ipic = Polar2D.get_fig(np.abs(last_reciprocal_intensity),transparent_backgound=True,colormap='inferno',scale='log', print_colorscale=True,n_pixels=512)
        #log.info('ipic shape {} ipic type {}'.format(ipic.shape,ipic.dtype))
        #log.info('dpic shape {} dpic type {}'.format(dpic.shape,dpic.dtype))
        out_pic = np.concatenate((dpic,ipic),axis=1)
        #log.info(pic_path)
        self.save(pic_path,out_pic)
    def load_ccd(self,name,**kwargs):
        options = self.files['ccd'].get('options',{})        
        _type=options['type']
        input_data=self.load(name,**{'skip_custom_methods': True,**kwargs})
        if (_type == 'legacy'):
            data=self.load_ccd_legacy(input_data)
        elif _type == 'direct':
            data = self.load_ccd_direct(input_data) 
        else:
            e = AssertionError('ccd loading type {} not specifierd or known!'.format(_type))
            log.error(e)
            raise e
        return data

    def load_ccd_legacy(self,data):
        out_dict={}
        if 'intra' in data:
            cc=data['intra']['ccf_2p_q1q2'].real
        else:
            cc=data['ccf_q1q2_2p'].real
        if cc.shape[0] < cc.shape[1]:            
            qs=data['q2']#/(2*np.pi)
            step_size = int(np.round(cc.shape[1]/cc.shape[0]))
            cc = cc[:,::step_size]
            a_int=data['iaverage'][::step_size]
        elif cc.shape[0] > cc.shape[1]:            
            qs=data['q1']#/(2*np.pi)
            step_size = np.round(cc.shape[0]/cc.shape[1])
            cc = cc[::step_size,:]
            a_int=data['iaverage'][::step_size]
        else:
            qs=data['q1']#/(2*np.pi)
            a_int=data['iaverage']

        other_cc = {}
        for key,name in zip(['ccf_q1q2_3p','ccf_q1q2_4p'],['I2I1','I2I2']):
            if key in data:
                other_cc[name] = data[key].real
        #cc[0,0]=np.mean(cc[0,0]) # test purposes
        out_dict['cross_correlation'] = {**{'I1I1':cc},**other_cc}
        #log.info('iaverage = {}'.format(a_int))
        out_dict['average_intensity']=a_int

        #qs[0] = 0 #test purposes
        out_dict['radial_points']=qs
        out_dict['qs']=qs
        #log.info('\n MAX Q={}'.format(qs[-1]))
        phis = data['phi']
        out_dict['angular_points']=phis
        out_dict['phis']=phis        
        out_dict['pi_in_q']=data.get('pi_in_q',True)
        
        q_phi_grid=GridFactory.construct_grid('uniform',[qs,phis])
        #            log.info('aInt shape={}'.format(aIntGrid.shape))
        #log.info('aInt grid dimension={}'.format(q_phi_grid.total_shape[-1]))
        #            log.info('first + last qs: \n{}  | {}'.format(q[:4],q[-4:]))
        #            log.info('first + last phis: \n{}  | {}'.format(phi[:4],phi[-4:]))
        #            stepLength={'q':q[2]-q[1],'phi':phi[2]-phi[1]}
        #            domains={'q':[q[0],q[-1]],'phi':[0,2*np.pi+stepLength['phi']]}
        out_dict['average_intensity'] = SampledFunction(NestedArray(qs[:,None],1),a_int,coord_sys='cartesian')
        out_dict['xray_wavelength']=data.get('xray_wavelength',1.23984) # in angström
        if out_dict['pi_in_q']:            
            thetas = ewald_sphere_theta_pi(out_dict['xray_wavelength'],qs)
        else:
            thetas = ewald_sphere_theta(out_dict['xray_wavelength'],qs)
        out_dict['thetas'] = thetas
        out_dict['data_grid'] = {'qs':qs,'thetas':thetas,'phis':phis}
        out_dict['dimensions'] = settings.project.dimensions
        return out_dict

    def load_ccd_direct(self,data):
        qs = data['radial_points']
        phis = data['angular_points']
        grid=GridFactory.construct_grid('uniform',[qs,phis])
        a_int = data['average_intensity']
        average_intensity = SampledFunction(NestedArray(grid[:,0,0],1),a_int,coord_sys='cartesian')
        data['average_intensity'] = average_intensity
        log.info(f'data = {data.keys()}')
        thetas = ewald_sphere_theta_pi(data['xray_wavelength'],qs)
        data['data_grid'] = {'qs':qs,'thetas':thetas,'phis':phis}
        data['dimensions'] = settings.project.dimensions
        #log.info(data['cross_correlation']['I2I2'])
        return data

    def load_invariants(self,name,**kwargs):
        data = self.load_direct(name,**kwargs)
        if isinstance(data['data_projection_matrices'],np.ndarray):
            matrices = data['data_projection_matrices']
        else:
            if 'I1I1' in data['data_projection_matrices']:
                matrices = data['data_projection_matrices']['I1I1']
                data['data_projection_matrices_2'] = data['data_projection_matrices']
                
            else:
                matrices = data['data_projection_matrices']
        low_res_matrices = data.get('data_low_resolution_intensity_coefficients',False)        
        a_int = data['average_intensity']
        r_pt = data['data_radial_points']        
        data['average_intensity'] = SampledFunction(NestedArray(r_pt[:,None],1),a_int,coord_sys='cartesian')
        b_coeff = data.get('deg_2_invariant',False)
        data['b_coeff'] = b_coeff
        if isinstance(matrices,dict):
            sorted_keys = np.sort(tuple(int(key) for key in matrices.keys())).astype(str)
            matrices = tuple(matrices[k] for k in sorted_keys)
        if data['dimensions'] == 3:
            #log.info('projection_matrices 0 shape={}'.format(matrices['0'].shape))
            #matrices['0']=matrices['0'][:,None]
            #log.info('projection_matrices 0 shape={}'.format(matrices['0'].shape))
            #
            
            tmp = np.empty(len(matrices),object)
            for i,pm in enumerate(matrices):
                tmp[i]=pm
            matrices = tmp
            data['data_projection_matrices']=matrices #tuple(matrices)
            
            if not isinstance(low_res_matrices,bool):
                log.info(f'low res mat type = {type(low_res_matrices)}')
                tmp = np.empty(len(low_res_matrices),object)
                for i,pm in enumerate(low_res_matrices):
                    tmp[i]=pm
                low_res_matrices = tmp
            if len(matrices[0].shape)<2:
                matrices[0]=matrices[0][:,None] #legacy compatibility
            data['data_low_resolution_intensity_coefficients']=low_res_matrices #tuple(matrices)
        elif data['dimensions'] == 2:
            data['data_projection_matrices'] = np.array(matrices)
        return data
    
    def save_invariants(self,name,proj_class,**options):
        #log.info(options.keys())
        time_str=self.get_time_string()
        path=self.folders[self.files['invariants']['folder']]
        path_modifiers = {'dimensions':settings.project.dimensions,'structure_name':settings.project.structure_name,'date':time_str,}
        run= self.get_latest_run('invariants',path_modifiers=path_modifiers) + 1
        path_modifiers['run'] = run
        run_path=path.format(**path_modifiers)
        
        data_dict={}
        data_dict['dimensions']=proj_class.dimensions
        data_dict['xray_wavelength'] = proj_class.xray_wavelength
        data_dict['average_intensity'] = proj_class.average_intensity[:]
        data_dict['data_radial_points'] = proj_class.data_radial_points
        data_dict['data_angular_points'] = proj_class.data_angular_points
        data_dict['data_min_q'] = proj_class.data_min_q
        #matrices = proj_class.data_projection_matrices        
        if proj_class.dimensions == 3:
            data_dict['data_projection_matrices'] = proj_class.data_projection_matrices
            data_dict['data_low_resolution_intensity_coefficients'] = proj_class.data_low_resolution_intensity_coefficients
            #for key,matrices in proj_class.data_projection_matrices.items():
            #    matrices = (np.squeeze(matrices[0]),)+tuple(matrices[1:])
            #    matrices = {str(key):value for key,value in enumerate(matrices)}
            #    data_dict['data_projection_matrices'][key] = matrices
        elif proj_class.dimensions == 2:
            #log.info('n proj matrices = {} dtype = {}'.format(len(matrices),matrices.dtype))
            data_dict['data_projection_matrices'] = proj_class.data_projection_matrices
            data_dict['data_projection_matrix_error_estimates'] = proj_class.data_projection_matrix_error_estimates
        data_dict['data_projection_matrices_q_id_limits']=proj_class.data_projection_matrices_q_id_limits
        data_dict['max_order'] = proj_class.max_order
        #data_dict['pi_in_q'] = int(proj_class.pi_in_q)
        data_dict['number_of_particles'] = int(proj_class.number_of_particles)
        if options.get('save_invariant',False):
            data_dict['deg_2_invariant'] = proj_class.b_coeff
            data_dict['deg_2_invariant_masks'] = proj_class.b_coeff_masks
            data_dict['deg_2_invariant_q_id_limits'] = proj_class.b_coeff_q_id_limits

        data_folder = self.get_path('invariants_archive',path_modifiers = path_modifiers,is_file=False)
        options['path_modifiers']=path_modifiers
        self.save_direct(name,data_dict,**options)
        if options.get('create_symlink',False):
            self.create_symlink(name,'invariant_symlink',path_modifiers=path_modifiers)
        
        file_name = 'extraction_settings'
        self._save_settings(data_folder,project_name=file_name)

        try:
            xprint('first invar')
            if options.get('plot_first_invariants',False):
                for key,bls in proj_class.b_coeff.items():
                    bl_args = np.angle(bls)+np.pi
                    mask = proj_class.b_coeff_masks[key]
                    self._save_first_invariants(bls,proj_class.data_radial_points,data_folder,options,name="first_{}_".format(key))
                    self._save_first_invariants(bl_args,proj_class.data_radial_points,data_folder,options,name="first_arg_of_{}_".format(key))
                    self._save_first_invariants(bls,proj_class.data_radial_points,data_folder,options,name="mask_of_{}_".format(key),mask = mask)
        except Exception as e:
            log.info('Plotting first invariants failed !')
            traceback.print_exc()
        try:
            log.info(f'trying to plot errors for {proj_class.data_projection_matrix_error_estimates.keys()}')
            if options.get('plot_first_projection_matrix_error_estimates',False):                
                for key,pr_err in proj_class.data_projection_matrix_error_estimates.items():
                    #log.info(f'plotting erros for {key}')
                    mask = proj_class.b_coeff_masks[key]
                    o = {'plot_range':[1e-10,1]}
                    self._save_first_invariants(pr_err,proj_class.data_radial_points,data_folder,o,name="first_projection_matrix_errors_{}_".format(key),mask = mask)
        except Exception as e:
            log.info('Plotting first projection matrix error estimates failed !')
            traceback.print_exc()
        try:
            if options.get('plot_first_ccn',False) and (proj_class.dimensions == 3):
                for key,bls in proj_class.b_coeff.items():        
                    cns = deg2_invariant_to_cn_3d(bls,proj_class.data_radial_points,proj_class.xray_wavelength)
                    #log.info('bcoeff dtype = {}'.format(bls.dtype))
                    cns = np.moveaxis(cns,-1,0)
                    max_value = np.max(tuple(np.abs(cn).max() for cn in cns))
                    vmin,vmax = options.get('plot_range',[max_value*1e-12,max_value])            
                    if [vmin,vmax] == [None,None]:
                        vmin,vmax = [max_value*1e-12,max_value]
        
                    grid = GridFactory.construct_grid('uniform',[proj_class.data_radial_points,proj_class.data_radial_points])
                    order_ids = [0,2,4,6,8]
    
                    shape = [2,len(order_ids)]
                    layouts = []
                    for i in range(shape[0]):
                        layout_part = []
                        orders = np.arange(10*i,10*i+len(order_ids)*2,2)
                        for o in orders:
                            layout = {'title':'$B_{'+'{}'.format(o)+'}$',
                                      'x_label':'$q_1$',
                                      'y_label':'$q_2$'
                                      }
                            layout_part.append(layout)
                        layouts.append(layout_part)

            
                    fig_bl_masks = heat2D_multi.get_fig([[np.abs(cns[n]) for n in order_ids],[np.abs(cns[10+n]) for n in order_ids] ],scale = 'log',layout = layouts,grid =grid,shape = shape,size = (30,10),vmin= vmin, vmax = vmax,cmap='plasma')

                    bl_path = data_folder+'first_{}_CCn.matplotlib'.format(key)
                    self.save(bl_path,fig_bl_masks,dpi = 500)
        except Exception as e:
            log.info('Plotting first CCn failed !')
            traceback.print_exc()
            
        try:
            xprint('first invar proj')
            #log.info(f"trying to plot pr inv = {options.get('plot_first_invariants_from_proj_matrices',False)}")
            if options.get('plot_first_invariants_from_proj_matrices',False):
                #log.info('plot first invariants from projection matrices')
                bls2 = {}
                if proj_class.dimensions == 2:
                    for key,prs in proj_class.data_projection_matrices.items(): 
                        bls2[key] = harmonic_coeff_to_deg2_invariants(proj_class.dimensions,np.array(prs).T)
                elif proj_class.dimensions == 3:
                    for key,prs in proj_class.data_projection_matrices.items(): 
                        bls2[key] = harmonic_coeff_to_deg2_invariants(proj_class.dimensions,prs)
                for key,bls in bls2.items():
                    if not key == 'I2I1':
                        #log.info(f'bls shape = {bls.shape}')
                        self._save_first_invariants(bls,proj_class.data_radial_points,data_folder,options,name="first_{}_proj_matrices_to_".format(key))
        except Exception as e:
            log.info("Plotting Bl from projection matrices failed!")
            traceback.print_exc()
        try:            
            if options.get('save_intensity_vtk',False):
                if proj_class.dimensions == 3:
                    V = proj_class.data_projection_matrices['I1I1']
                    cht = HarmonicTransform('complex',{'dimensions':3,'n_phis':0,'n_thetas':0,'max_order':proj_class.max_order})
                    grid = GridFactory.construct_grid('uniform',[proj_class.data_radial_points,cht.grid_param['thetas'],cht.grid_param['phis']])
                    log.info(f'len proj = {len(V)}')
                    log.info(f'shape proj 2 = {V[2].shape}')
                    log.info(f'grid shape = ')
                    I = np.abs(cht.inverse(V))

                    path = run_path + 'intensity_guess.vts'
                    log.info(f" intensity to {path}")
                    self.save(path,[I.real],grid_type='spherical',grid = grid)
                    
        except Exception as e:
            log.info('Plotting Best Intensity guess failed !')
            traceback.print_exc()


    def save_ccd(self,name,data,**options):
        log.info('custom saving of cross correaltion')
        opt=settings.project
        time_str=self.get_time_string()
        ccd_folder = self.folders['ccd_archive']
        path_modifiers =  {'structure_name':opt.structure_name,'date':time_str}
        run = self.get_latest_run(ccd_folder,path_modifiers = path_modifiers)+1
        path_modifiers['run']=run
        log.info(f'path modifiers = {path_modifiers}')
        try:
            data_path = self.get_path(name,path_modifiers=path_modifiers)
            log.info(f'data_path = {data_path},name = {name}')
            self.save_direct(name,data,path_modifiers=path_modifiers)
            log.info('saved h5')
        except Exception as e:
            log.warning(f'Failed to save cross-correlation h5 data! with error {e}')
            log.info(traceback.format_exc())
            
        try:
            log.info('saving settings')
            cc_path = self.get_path(name,path_modifiers = path_modifiers)
            cc_folder = os.path.dirname(cc_path)
            cc_name = os.path.basename(cc_path).split('.')[0]
            log.info(f'to {cc_folder}')
            self._save_settings(cc_folder,project_name='settings')
        except Exception as e:
            log.warning(f'Failed to save cross-correlation settings! with error {e}')
            log.info(traceback.format_exc())
            
        if options.get('save_model_vtk',False):
            try:
                density = options['model_density']
                grid = options['grid']
                if grid.shape[-1]==3:
                    grid_type = 'spherical'
                elif grid.shape[-1]==2:
                    grid_type = 'polar'
                self.save('model_density',[density],dset_names=['model_density'],grid=grid,grid_type=grid_type,path_modifiers = path_modifiers)
            except Exception as e:
                log.warning(f'Failed to save model density! with error {e}')
                log.info(traceback.format_exc())
        if options.get('save_symlink',False):
            try:
                self.create_symlink(name,'ccd_symlink',path_modifiers=path_modifiers)
            except Exception as e:
                log.warning(f'Failed to create cross-corelation symlink! with error {e}')
                log.info(traceback.format_exc())
